# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT


from .resource import *
from .util import _Throttler


class KivyPropertyBacked(SenmlResource):
    """Provides a .value bound to the self.widget_property property of the
    self.backend_widget Kivy widget. The widget and widget_property need to be
    set before this class's __init__ is called."""

    def __init__(self):
        super().__init__()
        throttler = _Throttler(self.value_changed)
        self.backend_widget.bind(**{self.widget_property: lambda *args: throttler()})

    def _get_value(self):
        return getattr(self.backend_widget, self.widget_property)

    def _set_value(self, updated):
        setattr(self.backend_widget, self.widget_property, updated)

    value = property(_get_value, _set_value)


class Color(KivyPropertyBacked):
    def __init__(self, kivy_backend, widget_property):
        self.backend_widget = kivy_backend
        self.widget_property = widget_property
        super().__init__()

    @ContenttypeRendered.get_handler("text/plain;charset=utf-8", default=True)
    def __regular_get(self):
        return "#" + "".join("%02x" % int(255 * c) for c in self._get_value()[:3])

    # 65362 is the content format used for SAUL RGB values, also in verdigris (manual clients will just send none)
    #
    # See also https://rustdoc.etonomy.org/riot_coap_handler_demos/saul/index.html
    @ContenttypeRendered.put_handler(65362, default=True)
    def render_put(self, payload):
        if len(payload) == 7 and payload[0:1] == b"#":
            values = tuple(
                int(payload[i : i + 2].decode("ascii"), 16) / 255 for i in (1, 3, 5)
            )
        else:
            return Message(code=BAD_REQUEST)

        self._set_value(values + (1,))

    def get_link_description(self):
        return {"saul": "ACT_LED_RGB", "if": "core.p"}


class Text(StringResource, KivyPropertyBacked):
    """A resource that represents a Kivy widget's text"""

    if_ = "core.p"

    widget_property = "text"

    def __init__(self, kivy_backend):
        self.backend_widget = kivy_backend
        super().__init__()


class PressState(BooleanResource, KivyPropertyBacked):
    """A resource that resesents a Kivy widget's "pressed" property"""

    if_ = "core.s"

    widget_property = "state"

    # This is no good modelling (for a button modelled as press-release is hard
    # to observe with eventualy consistent semantics only), but it works for a
    # first demo.

    def __init__(self, kivy_backend):
        self.backend_widget = kivy_backend
        super().__init__()

    # fake boolean, and don't allow setting

    def _get_value(self):
        return {"down": 1, "normal": 0}[super()._get_value()]

    value = property(_get_value)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Helpers around content types

This uses the terminology clarified in 1_, and primarily deals with content
types in their usual string representation.

Unless content types get used a lot more in aiocoap, this provides only
accessors to some of their relevant properties, without aiming to build
semantically accessible objects to encapsulate them.

.. _1: https://tools.ietf.org/html/draft-bormann-core-media-content-type-format-01"""


def categorize(contenttype: str):
    """Return 'cbor', 'json' or 'link-format' if the content type indicates it
    is that format itself or derived from it."""

    media_type, *_ = contenttype.split(";")
    _, _, subtype = media_type.partition("/")

    if subtype == "cbor-seq" or subtype.endswith("+cbor-seq"):
        return "cbor-seq"

    if subtype == "cbor" or subtype.endswith("+cbor"):
        return "cbor"

    if subtype == "json" or subtype.endswith("+json"):
        return "json"

    if media_type == "application/link-format":
        return "link-format"

    return None



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This is a usage example of aiocoap that demonstrates how to implement a
simple client. See the "Usage Examples" section in the aiocoap documentation
for some more information."""

import logging
import asyncio

from aiocoap import *

logging.basicConfig(level=logging.INFO)


async def main():
    """Perform a single PUT request to localhost on the default port, URI
    "/other/block". The request is sent 2 seconds after initialization.

    The payload is bigger than 1kB, and thus sent as several blocks."""

    context = await Context.create_client_context()

    await asyncio.sleep(2)

    payload = b"The quick brown fox jumps over the lazy dog.\n" * 30
    request = Message(code=PUT, payload=payload, uri="coap://localhost/other/block")

    response = await context.request(request).response

    print("Result: %s\n%r" % (response.code, response.payload))


if __name__ == "__main__":
    asyncio.run(main())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Run the OSCORE plug test"""

import asyncio
import unittest
import tempfile
import shutil

import aiocoap
import aiocoap.defaults
from aiocoap.util import hostportjoin

from .test_server import WithAsyncLoop, WithClient, asynctest
from .fixtures import is_test_successful

from .common import PYTHON_PREFIX, CapturingSubprocess

SERVER_ADDRESS = "::1"
SERVER = PYTHON_PREFIX + [
    "./contrib/oscore-plugtest/plugtest-server",
    "--verbose",
    "--bind",
    hostportjoin(SERVER_ADDRESS, None),
]
CLIENT = PYTHON_PREFIX + ["./contrib/oscore-plugtest/plugtest-client", "--verbose"]

# those are to be expected to contain bad words -- 'Check passed: X failed' is legitimate
output_whitelist = ["Check passed: "]
# explicitly whitelisted for when the server is run with increased verbosity
debug_whitelist = [
    "INFO:coap-server:Render request raised a renderable error",
    "DEBUG:oscore-site:Will encrypt message as response: ",
    "DEBUG:aiocoap.cryptography:Unprotecting failed",
]


class WithAssertNofaillines(unittest.TestCase):
    def assertNoFaillines(self, text_to_check, message):
        """Assert that there are no lines that contain the phrase 'fail' or
        'WARNING'/'ERROR' in the output, unless they are a 'Check passed' line
        or other whitelisted ones.

        This is to check the output of the plugtest client, which may
        successfully report: 'Check passed: The validation failed. (Tag
        invalid)'"""

        lines = text_to_check.decode("utf8").split("\n")
        lines = (
            l
            # "failed" and "error" are always legitimate in this position
            # as they happen by design; whereever they are unexpected,
            # they're caught by the regular plug test operation
            .replace("Precondition Failed", "Precondition @@@led").replace(
                "Internal Server Error", "Internal Server @@@or"
            )
            for l in lines
        )
        lines = (
            l
            for l in lines
            if not any(l.startswith(white) for white in output_whitelist)
        )
        lines = (l for l in lines if not any(white in l for white in debug_whitelist))
        errorlines = (
            l
            for l in lines
            if "fail" in l.lower() or "warning" in l.lower() or "error" in l.lower()
        )
        self.assertEqual([], list(errorlines), message)


@unittest.skipIf(
    aiocoap.defaults.oscore_missing_modules(),
    "Module missing for running OSCORE tests: %s"
    % (aiocoap.defaults.oscore_missing_modules(),),
)
class WithPlugtestServer(WithAsyncLoop, WithAssertNofaillines):
    def setUp(self):
        super(WithPlugtestServer, self).setUp()
        ready = self.loop.create_future()
        self.__done = self.loop.create_future()

        self.contextdir = tempfile.mkdtemp(suffix="-contexts")

        self.__task = self.loop.create_task(self.run_server(ready, self.__done))
        self.__task.add_done_callback(
            lambda _: None
            if ready.done()
            else ready.set_exception(self.__task.exception())
        )
        self.loop.run_until_complete(ready)

    async def run_server(self, readiness, done):
        self.process, process_outputs = await self.loop.subprocess_exec(
            CapturingSubprocess, *self.SERVER, self.contextdir + "/server", stdin=None
        )
        try:
            while True:
                if b"Plugtest server ready.\n" in process_outputs.stdout:
                    break
                if self.process.get_returncode() is not None:
                    readiness.set_exception(
                        RuntimeError(
                            "OSCORE server process terminated during startup:\n%s\n%s"
                            % (
                                process_outputs.stdout.decode("utf8"),
                                process_outputs.stderr.decode("utf8"),
                            )
                        )
                    )
                    return
                await process_outputs.read_more
            readiness.set_result(True)

            while True:
                if self.process.get_returncode() is not None:
                    break
                await process_outputs.read_more

            done.set_result((process_outputs.stdout, process_outputs.stderr))

        finally:
            self.process.close()

    def tearDown(self):
        # Don't leave this over, even if anything is raised during teardown
        self.process.terminate()

        super().tearDown()

        out, err = self.loop.run_until_complete(self.__done)

        if not is_test_successful(self):
            if not out and not err:
                return
            self.fail(
                "Previous errors occurred."
                + (
                    "\nServer stdout was:\n    "
                    + out.decode("utf8").replace("\n", "\n    ")
                    if out
                    else ""
                )
                + (
                    "\nServer stderr was:\n    "
                    + err.decode("utf8").replace("\n", "\n    ")
                    if err
                    else ""
                )
            )
        else:
            self.assertNoFaillines(out, '"failed" showed up in plugtest server stdout')
            self.assertNoFaillines(err, '"failed" showed up in plugtest server stderr')

        # Unlike the server process termination, leaving those around can be
        # helpful and barely does any harm.
        shutil.rmtree(self.contextdir)


class TestOSCOREPlugtestBase(WithPlugtestServer, WithClient, WithAssertNofaillines):
    @asynctest
    async def _test_plugtestclient(self, x):
        proc, transport = await self.loop.subprocess_exec(
            CapturingSubprocess,
            *(
                CLIENT
                + ["[" + SERVER_ADDRESS + "]", self.contextdir + "/client", str(x)]
            ),
            stdin=None,
        )

        try:
            while True:
                if proc.get_returncode() is not None:
                    break
                await transport.read_more
        except asyncio.CancelledError:
            proc.terminate()
        else:
            proc.close()

        self.assertEqual(
            proc.get_returncode(),
            0,
            "Plugtest client return non-zero exit state\nOutput was:\n"
            + transport.stdout.decode("utf8")
            + "\nErrorr output was:\n"
            + transport.stderr.decode("utf8"),
        )
        self.assertNoFaillines(
            transport.stdout, '"failed" showed up in plugtest client stdout'
        )
        self.assertNoFaillines(
            transport.stderr, '"failed" showed up in plugtest client stderr'
        )


class TestOSCOREPlugtestWithoutRecovery(TestOSCOREPlugtestBase):
    SERVER = SERVER


class TestOSCOREPlugtestWithRecovery(TestOSCOREPlugtestBase):
    SERVER = SERVER + ["--state-was-lost"]


for x in range(0, 17):
    for cls in (TestOSCOREPlugtestWithRecovery, TestOSCOREPlugtestWithoutRecovery):
        t = lambda self, x=x: self._test_plugtestclient(x)
        if x == 16:
            # That test can not succeed against a regular plugtest server
            t = unittest.expectedFailure(t)
        if x == 7:
            # That test fails because there is no proper observation cancellation
            # aroun yet, see https://github.com/chrysn/aiocoap/issues/104
            #
            # Not making a statement on whether this is ecpected to work or
            # not, because it is highly irregular (it works with setup.py test
            # and fails with tox?)
            continue

        # enforcing them to sort properly is purely a readability thing, they
        # execute correctly out-of-order too.
        setattr(cls, "test_%03d" % x, t)
    # Let's not leak a global that'd be picked up for testing, given these are
    # already being tested
    del cls



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""aiocoap-client is a simple command-line tool for interacting with CoAP servers"""

import sys
import asyncio
import argparse
import logging
import subprocess
from pathlib import Path

import shlex

# even though not used directly, this has side effects on the input() function
# used in interactive mode
try:
    import readline  # noqa: F401
except ImportError:
    pass  # that's normal on some platforms, and ok since it's just a usability enhancement

import aiocoap
import aiocoap.defaults
import aiocoap.meta
import aiocoap.proxy.client
from aiocoap.util import contenttype
from aiocoap.util.cli import ActionNoYes
from aiocoap.numbers import ContentFormat


def build_parser():
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument(
        "--non",
        help="Send request as non-confirmable (NON) message",
        action="store_true",
    )
    p.add_argument(
        "-m",
        "--method",
        help="Name or number of request method to use (default: %(default)s)",
        default="GET",
    )
    p.add_argument(
        "--observe", help="Register an observation on the resource", action="store_true"
    )
    p.add_argument(
        "--observe-exec",
        help="Run the specified program whenever the observed resource changes, feeding the response data to its stdin",
        metavar="CMD",
    )
    p.add_argument(
        "--accept",
        help="Content format to request",
        metavar="MIME",
    )
    p.add_argument(
        "--proxy", help="Relay the CoAP request to a proxy for execution", metavar="URI"
    )
    p.add_argument(
        "--payload",
        help="Send X as request payload (eg. with a PUT). If X starts with an '@', its remainder is treated as a file name and read from; '@-' reads from the console. Non-file data may be recoded, see --content-format.",
        metavar="X",
    )
    p.add_argument(
        "--payload-initial-szx",
        help="Size exponent to limit the initial block's size (0 ≙ 16 Byte, 6 ≙ 1024 Byte)",
        metavar="SZX",
        type=int,
    )
    p.add_argument(
        "--content-format",
        help="Content format of the --payload data. If a known format is given and --payload has a non-file argument, the payload is converted from CBOR Diagnostic Notation.",
        metavar="MIME",
    )
    p.add_argument(
        "--no-set-hostname",
        help="Suppress transmission of Uri-Host even if the host name is not an IP literal",
        dest="set_hostname",
        action="store_false",
        default=True,
    )
    p.add_argument(
        "-v",
        "--verbose",
        help="Increase the debug output",
        action="count",
    )
    p.add_argument(
        "-q",
        "--quiet",
        help="Decrease the debug output",
        action="count",
    )
    # careful: picked before parsing
    p.add_argument(
        "--interactive",
        help="Enter interactive mode",
        action="store_true",
    )
    p.add_argument(
        "--credentials",
        help="Load credentials to use from a given file",
        type=Path,
    )
    p.add_argument(
        "--version", action="version", version="%(prog)s " + aiocoap.meta.version
    )

    p.add_argument(
        "--color",
        help="Color output (default on TTYs if all required modules are installed)",
        default=None,
        action=ActionNoYes,
    )
    p.add_argument(
        "--pretty-print",
        help="Pretty-print known content formats (default on TTYs if all required modules are installed)",
        default=None,
        action=ActionNoYes,
    )
    p.add_argument(
        "url",
        help="CoAP address to fetch",
    )

    return p


def configure_logging(verbosity):
    logging.basicConfig()

    if verbosity <= -2:
        logging.getLogger("coap").setLevel(logging.CRITICAL + 1)
    elif verbosity == -1:
        logging.getLogger("coap").setLevel(logging.ERROR)
    elif verbosity == 0:
        logging.getLogger("coap").setLevel(logging.WARNING)
    elif verbosity == 1:
        logging.getLogger("coap").setLevel(logging.INFO)
    elif verbosity >= 2:
        logging.getLogger("coap").setLevel(logging.DEBUG)


def colored(text, options, tokenlambda):
    """Apply pygments based coloring if options.color is set. Tokelambda is a
    callback to which pygments.token is passed and which returns a token type;
    this makes it easy to not need to conditionally react to pygments' possible
    absence in all color locations."""
    if not options.color:
        return str(text)

    from pygments.formatters import TerminalFormatter
    from pygments import token, format

    return format(
        [(tokenlambda(token), str(text))],
        TerminalFormatter(),
    )


def incoming_observation(options, response):
    if options.observe_exec:
        p = subprocess.Popen(options.observe_exec, shell=True, stdin=subprocess.PIPE)
        # FIXME this blocks
        p.communicate(response.payload)
    else:
        sys.stdout.write(colored("---", options, lambda token: token.Comment.Preproc))
        if response.code.is_successful():
            present(response, options, file=sys.stderr)
        else:
            sys.stdout.flush()
            print(
                colored(
                    response.code, options, lambda token: token.Token.Generic.Error
                ),
                file=sys.stderr,
            )
            if response.payload:
                present(response, options, file=sys.stderr)


def apply_credentials(context, credentials, errfn):
    if credentials.suffix == ".json":
        import json

        context.client_credentials.load_from_dict(json.load(credentials.open("rb")))
    elif credentials.suffix == ".diag":
        try:
            import cbor_diag
            import cbor2
        except ImportError:
            raise errfn(
                "Loading credentials in CBOR diagnostic format requires cbor2 and cbor_diag package"
            )
        context.client_credentials.load_from_dict(
            cbor2.loads(cbor_diag.diag2cbor(credentials.open().read()))
        )
    else:
        raise errfn(
            "Unknown suffix: %s (expected: .json or .diag)" % (credentials.suffix)
        )


def present(message, options, file=sys.stdout):
    """Write a message payload to the output, pretty printing and/or coloring
    it as configured in the options."""
    if not options.quiet and (message.opt.location_path or message.opt.location_query):
        # FIXME: Percent encoding is completely missing; this would be done
        # most easily with a CRI library
        location_ref = "/" + "/".join(message.opt.location_path)
        if message.opt.location_query:
            location_ref += "?" + "&".join(message.opt.location_query)
        print(
            colored(
                f"Location options indicate new resource: {location_ref}",
                options,
                lambda token: token.Token.Generic.Inserted,
            ),
            file=sys.stderr,
        )

    if not message.payload:
        return

    payload = None

    cf = message.opt.content_format or message.request.opt.content_format
    if cf is not None and cf.is_known():
        mime = cf.media_type
    else:
        mime = "application/octet-stream"
    if options.pretty_print:
        from aiocoap.util.prettyprint import pretty_print

        prettyprinted = pretty_print(message)
        if prettyprinted is not None:
            (infos, mime, payload) = prettyprinted
            if not options.quiet:
                for i in infos:
                    print(
                        colored("# " + i, options, lambda token: token.Comment),
                        file=sys.stderr,
                    )

    color = options.color
    if color:
        from aiocoap.util.prettyprint import lexer_for_mime
        import pygments

        try:
            lexer = lexer_for_mime(mime)
        except pygments.util.ClassNotFound:
            color = False

    if color and payload is None:
        # Coloring requires a unicode-string style payload, either from the
        # mime type or from the pretty printer.
        try:
            payload = message.payload.decode("utf8")
        except UnicodeDecodeError:
            color = False

    if color:
        from pygments.formatters import TerminalFormatter
        from pygments import highlight

        highlit = highlight(
            payload,
            lexer,
            TerminalFormatter(),
        )
        # The TerminalFormatter already adds an end-of-line character, not
        # trying to add one for any missing trailing newlines.
        print(highlit, file=file, end="")
        file.flush()
    else:
        if payload is None:
            file.buffer.write(message.payload)
            if file.isatty() and message.payload[-1:] != b"\n":
                file.write("\n")
        else:
            file.write(payload)
            if file.isatty() and payload[-1] != "\n":
                file.write("\n")


async def single_request(args, context):
    parser = build_parser()
    options = parser.parse_args(args)

    pretty_print_modules = aiocoap.defaults.prettyprint_missing_modules()
    if pretty_print_modules and (options.color is True or options.pretty_print is True):
        parser.error(
            "Color and pretty printing require the following"
            " additional module(s) to be installed: %s"
            % ", ".join(pretty_print_modules)
        )
    if options.color is None:
        options.color = sys.stdout.isatty() and not pretty_print_modules
    if options.pretty_print is None:
        options.pretty_print = sys.stdout.isatty() and not pretty_print_modules

    configure_logging((options.verbose or 0) - (options.quiet or 0))

    try:
        try:
            code = getattr(
                aiocoap.numbers.codes.Code,
                options.method.upper().replace("IPATCH", "iPATCH"),
            )
        except AttributeError:
            try:
                code = aiocoap.numbers.codes.Code(int(options.method))
            except ValueError:
                raise parser.error("Unknown method")

        if options.credentials is not None:
            apply_credentials(context, options.credentials, parser.error)

        request = aiocoap.Message(
            code=code, mtype=aiocoap.NON if options.non else aiocoap.CON
        )
        request.set_request_uri(options.url, set_uri_host=options.set_hostname)

        if options.accept:
            try:
                request.opt.accept = ContentFormat(int(options.accept))
            except ValueError:
                try:
                    request.opt.accept = ContentFormat.by_media_type(options.accept)
                except KeyError:
                    raise parser.error("Unknown accept type")

        if options.observe:
            request.opt.observe = 0
            observation_is_over = asyncio.get_event_loop().create_future()

        if options.content_format:
            try:
                request.opt.content_format = ContentFormat(int(options.content_format))
            except ValueError:
                try:
                    request.opt.content_format = ContentFormat.by_media_type(
                        options.content_format
                    )
                except KeyError:
                    raise parser.error("Unknown content format")

        if options.payload:
            if options.payload.startswith("@"):
                filename = options.payload[1:]
                if filename == "-":
                    f = sys.stdin.buffer
                else:
                    f = open(filename, "rb")
                try:
                    request.payload = f.read()
                except OSError as e:
                    raise parser.error("File could not be opened: %s" % e)
            else:
                request_classification = contenttype.categorize(
                    request.opt.content_format.media_type
                    if request.opt.content_format is not None
                    and request.opt.content_format.is_known()
                    else ""
                )
                if request_classification in ("cbor", "cbor-seq"):
                    try:
                        import cbor_diag
                    except ImportError as e:
                        raise parser.error(f"CBOR recoding not available ({e})")

                    try:
                        encoded = cbor_diag.diag2cbor(options.payload)
                    except ValueError as e:
                        raise parser.error(
                            f"Parsing CBOR diagnostic notation failed. Make sure quotation marks are escaped from the shell. Error: {e}"
                        )

                    if request_classification == "cbor-seq":
                        try:
                            import cbor2
                        except ImportError as e:
                            raise parser.error(
                                f"CBOR sequence recoding not available ({e})"
                            )
                        decoded = cbor2.loads(encoded)
                        if not isinstance(decoded, list):
                            raise parser.error(
                                "CBOR sequence recoding requires an array as the top-level element."
                            )
                        request.payload = b"".join(cbor2.dumps(d) for d in decoded)
                    else:
                        request.payload = encoded
                else:
                    request.payload = options.payload.encode("utf8")

        if options.payload_initial_szx is not None:
            request.remote.maximum_block_size_exp = options.payload_initial_szx

        if options.proxy is None:
            interface = context
        else:
            interface = aiocoap.proxy.client.ProxyForwarder(options.proxy, context)

        requested_uri = request.get_request_uri()

        requester = interface.request(request)

        if options.observe:
            requester.observation.register_errback(observation_is_over.set_result)
            requester.observation.register_callback(
                lambda data, options=options: incoming_observation(options, data)
            )

        try:
            response_data = await requester.response
        finally:
            if not requester.response.done():
                requester.response.cancel()
            if options.observe and not requester.observation.cancelled:
                requester.observation.cancel()

        response_uri = response_data.get_request_uri()
        if requested_uri != response_uri:
            print(
                colored(
                    f"Response arrived from different address; base URI is {response_uri}",
                    options,
                    lambda token: token.Generic.Inserted,
                ),
                file=sys.stderr,
            )
        if response_data.code.is_successful():
            present(response_data, options)
        else:
            print(
                colored(response_data.code, options, lambda token: token.Generic.Error),
                file=sys.stderr,
            )
            present(response_data, options, file=sys.stderr)
            sys.exit(1)

        if options.observe:
            exit_reason = await observation_is_over
            print("Observation is over: %r" % (exit_reason,), file=sys.stderr)
    except aiocoap.error.HelpfulError as e:
        print(str(e), file=sys.stderr)
        extra_help = e.extra_help(
            hints=dict(
                original_uri=options.url,
                request=request,
            )
        )
        if extra_help:
            print("Debugging hint:", extra_help, file=sys.stderr)
        sys.exit(1)
    # Fallback while not all backends raise NetworkErrors
    except OSError as e:
        text = str(e)
        if not text:
            text = repr(e)
        if not text:
            # eg ConnectionResetError flying out of a misconfigured SSL server
            text = type(e)
        print(
            "Warning: OS errors should not be raised this way any more.",
            file=sys.stderr,
        )
        # not telling what to do precisely: the form already tells users to
        # include `aiocoap.cli.defaults` output, which is exactly what we
        # need.
        print(
            f"Even if the cause of the error itself is clear, please file an issue at {aiocoap.meta.bugreport_uri}.",
            file=sys.stderr,
        )
        print("Error:", text, file=sys.stderr)
        sys.exit(1)


async def single_request_with_context(args):
    """Wrapper around single_request until sync_main gets made fully async, and
    async context managers are used to manage contexts."""
    context = await aiocoap.Context.create_client_context()
    try:
        await single_request(args, context)
    finally:
        await context.shutdown()


interactive_expecting_keyboard_interrupt = None


async def interactive():
    global interactive_expecting_keyboard_interrupt
    interactive_expecting_keyboard_interrupt = asyncio.get_event_loop().create_future()

    context = await aiocoap.Context.create_client_context()

    while True:
        try:
            # when http://bugs.python.org/issue22412 is resolved, use that instead
            line = await asyncio.get_event_loop().run_in_executor(
                None, lambda: input("aiocoap> ")
            )
        except EOFError:
            line = "exit"
        line = shlex.split(line)
        if not line:
            continue
        if line in (["help"], ["?"]):
            line = ["--help"]
        if line in (["quit"], ["q"], ["exit"]):
            break

        current_task = asyncio.create_task(
            single_request(line, context=context),
            name="Interactive prompt command %r" % line,
        )
        interactive_expecting_keyboard_interrupt = (
            asyncio.get_event_loop().create_future()
        )

        done, pending = await asyncio.wait(
            [current_task, interactive_expecting_keyboard_interrupt],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if current_task not in done:
            current_task.cancel()
        else:
            try:
                await current_task
            except SystemExit as e:
                if e.code != 0:
                    print("Exit code: %d" % e.code, file=sys.stderr)
                continue
            except Exception as e:
                print("Unhandled exception raised: %s" % (e,))

    await context.shutdown()


def sync_main(args=None):
    # interactive mode is a little messy, that's why this is not using aiocoap.util.cli yet
    if args is None:
        args = sys.argv[1:]

    if "--interactive" not in args:
        try:
            asyncio.run(single_request_with_context(args))
        except KeyboardInterrupt:
            sys.exit(3)
    else:
        if len(args) != 1:
            print(
                "No other arguments must be specified when entering interactive mode",
                file=sys.stderr,
            )
            sys.exit(1)

        loop = asyncio.get_event_loop()
        task = loop.create_task(
            interactive(),
            name="Interactive prompt",
        )

        while not task.done():
            try:
                loop.run_until_complete(task)
            except KeyboardInterrupt:
                if not interactive_expecting_keyboard_interrupt.done():
                    interactive_expecting_keyboard_interrupt.set_result(None)
            except SystemExit:
                continue  # asyncio/tasks.py(242) raises those after setting them as results, but we particularly want them back in the loop


if __name__ == "__main__":
    sync_main()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import unittest

from .fixtures import asynctest, no_warnings, WithAsyncLoop, WithLogMonitoring
from . import common

from aiocoap import Context


class TestProtocolSetup(WithLogMonitoring, WithAsyncLoop):
    """Tests that are only concerned with the setup of contexts, and the way
    they can be set up."""

    @no_warnings
    @asynctest
    async def test_empty_setup_shutdown(self):
        ctx = await Context.create_client_context()
        await ctx.shutdown()

    # There is not yet a way to set things up in an async context manager.

    #     @no_warnings
    #     @asynctest
    #     async def test_empty_contextmgr(self):
    #         async with Context.create_client_context():
    #             pass

    # The following tests should be converted to context managers once usable.

    @no_warnings
    @asynctest
    # Workaround for https://github.com/chrysn/aiocoap/issues/321
    @unittest.skipIf(
        hasattr(common, "gbulb"),
        reason="uvloop has unresolved issues with unused contexts",
    )
    async def test_multiple_contexts(self):
        # Not that that'd be a regular thing to do, just checking it *can* be
        # done
        c1 = await Context.create_client_context(loggername="coap-client1")
        c2 = await Context.create_client_context(loggername="coap-client2")
        # None is an acceptable site; binding to a concrete port
        # removes the worries of situations where the default
        # transports can't bind to "any".
        s1 = await Context.create_server_context(
            None, bind=("::1", None), loggername="coap-server"
        )

        await c1.shutdown()
        await s1.shutdown()
        await c2.shutdown()

    @no_warnings
    @asynctest
    async def test_serverports_no_conflict(self):
        # When different ports are used, servers should not get into conflict.
        #
        # (To some extent, that this is so easy is the fault of the weird way
        # the other protocols' ports are set for lack of configurability).
        s1 = await Context.create_server_context(None, bind=("::1", 1234))
        s2 = await Context.create_server_context(None, bind=("::1", None))

        await s1.shutdown()
        await s2.shutdown()
